/*
 QueryifyJs is a No-SQL object query library made for javascript
 Copyright (C) 2014 Peter Flannery

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @module QueryifyJs
 */
/**
 * Class for creating synchronous Query instances.
 *
 * @class Query
 *
 * @constructor
 * @param {QueryOperations} queryOperations
 * @param {QueryComparisons} queryComparisons
 * @param {QueryStackTrace} queryStackTrace
 * @param {Object} sharedContext
 *      passed to the custom compare methods
 */
function Query(queryOperations, queryComparisons, queryStackTrace, sharedContext) {
  this.opMethods = queryOperations;
  this.compareMethods = queryComparisons;
  this.trace = queryStackTrace;

  // fallback to an empty object if not supplied
  this.sharedContext = sharedContext || {};
}

Query.prototype = {
  /**
   * Extracts items from an object dictionary using a criteria object
   *
   * @method extract
   * @async
   *
   * @param {Object} modelDictionary An object dictionary of model objects
   * @param {Object} criteria The criteria object
   *
   * @return an {Object} dictionary of model objects.
   *
   * @example
   *
   * ```javascript
   *   // the data to query
   *   var modelDictionary = {
   *     "item1": {
   *       name: 'cyborg',
   *       age: 123
   *     },
   *     "item2": {
   *       name: 'borg',
   *       age: 223
   *     }
   *   };
   *
   *   // the criteria
   *   var criteria = {
   *     name: {
   *       $equals: 'cyborg'
   *     }
   *   };
   *   // run the query
   *   var results = queryify.extract(modelDictionary, criteria)
   *   // the results, an extracted model dictionary
   * ```
   */
  extract: function(modelDictionary, criteria) {
    var trace = this.trace.on,
      filteredDictionary = {},
      keyIndex,
      modelKey,
      model;

    if (trace === true) {
      // setup the query logging stack
      this.trace.reset();
    }

    if (modelDictionary === undefined) {
      throw new ReferenceError("Unspecified model dictionary.");
    }

    if (criteria === undefined) {
      throw new ReferenceError("Unspecified criteria.");
    }

    var modelDictKeys = Object.keys(modelDictionary);
    for (keyIndex = 0; keyIndex < modelDictKeys.length; keyIndex++) {
      if (trace === true) {
        this.trace.stackLevel = 0;
        this.trace.write(["$$extract", modelKey]);
      }

      modelKey = modelDictKeys[keyIndex];
      model = modelDictionary[modelKey];
      var result = this.opMethods.executeQuery.call(this, model, criteria, false);
      // determine if passed
      if (result === true) {
        filteredDictionary[modelKey] = model;
      }
    }

    return filteredDictionary;
  },
  /**
   * Filters an array of model objects against a criteria object
   *
   * @method filter
   * @async
   *
   * @param {Array} modelArray An array of data objects
   * @param {Object} criteria The criteria object
   *
   * @return {Promise} who's result is set to an {Array} of filtered objects.
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = [{
   *    name: 'cyborg',
   *    age: 123
   *  }, {
   *    name: 'borg',
   *    age: 223
   *  }];
   *
   *  // the criteria
   *  var criteria = {
   *    name: {
   *      $equals: 'cyborg'
   *    }
   *  };
   *
   *  // run the query
   *  var results = queryify.filter(model, criteria)
   *  // the results, a filtered array
   * ```
   */
  filter: function(modelArray, criteria) {
    var trace = this.trace.on;

    if (trace === true) {
      // setup the query logging stack
      this.trace.reset();
    }

    if (modelArray === undefined) {
      throw new ReferenceError("Unspecified model array.");
    }

    if (criteria === undefined) {
      throw new ReferenceError("Unspecified criteria.");
    }

    var result,
      model,
      traceIndex = 0,
      nextIter = modelArray[Symbol.iterator](),
      iterResult,
      filteredList = [];

    // iterates the promises until false or error
    for (; !(iterResult = nextIter.next()).done;) {
      if (trace === true) {
        // log current operation
        this.trace.stackLevel = 0;
        this.trace.write(["$$filter", traceIndex++]);
      }

      model = iterResult.value;
      result = this.opMethods.executeQuery.call(this, model, criteria, false);

      // determine if passed
      if (result === true) {
        filteredList.push(model);
      }
    }

    return filteredList;
  }
  ,
  /**
   * Tests if an array of model objects match against a criteria object
   *
   * @method every
   * @async
   *
   * @param {Array} modelArray An array of data objects
   * @param {Object} criteria The criteria object
   *
   * @return {Promise} who's result is set to a {Boolean}.
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = [{
   *    name: 'cyborg',
   *    age: 123
   *  }, {
   *    name: 'borg',
   *    age: 223
   *  }];
   *
   *  // the criteria
   *  var criteria = {
   *    name: {
   *      $right: 'g'
   *    },
   *    age: {
   *      $typeof: 'number'
   *    }
   *  };
   *
   *  // run the query
   *  var result = queryify.every(model, criteria)
   *  // the result, true or false
   * ```
   */
  every: function(modelArray, criteria) {
    var trace = this.trace.on;

    if (trace === true) {
      // setup the query logging stack
      this.trace.reset();
    }

    if (modelArray === undefined) {
      throw new ReferenceError("Unspecified model array.");
    }

    if (criteria === undefined) {
      throw new ReferenceError("Unspecified criteria.");
    }

    var result,
      traceIndex = 0,
      nextIter = modelArray[Symbol.iterator](),
      iterResult;

    // iterates the promises until false or error
    for (; !(iterResult = nextIter.next()).done;) {
      if (trace === true) {
        // log current item operation
        this.trace.stackLevel = 0;
        this.trace.write(["$$every", traceIndex++]);
      }

      result = this.opMethods.executeQuery.call(this, iterResult.value, criteria, false);
      if (result === false)
        break;
    }

    return result;
  },
  /**
   * Tests one model object against a criteria object
   *
   * @method one
   * @async
   *
   * @param {Object} model The data object
   * @param {Object} criteria The criteria object
   *
   * @return {Promise} who's result is set to a {Boolean}.
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *      name: 'cyborg',
   *      age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    name: {
   *      $right: 'g',
   *      $and: {
   *        $left: 'cy'
   *      }
   *    },
   *    $and: {
   *      age: {
   *        $typeof: 'number',
   *        $and: {
   *          $equals: 123
   *        }
   *      }
   *    }
   *  };
   *
   *  // run the query
   *  var result = queryify.one(model, criteria)
   *  // the result, true or false
   * ```
   */
  one: function(model, criteria) {
    if (this.trace.one === true) {
      // setup the query logging stack
      this.trace.reset();
    }

    if (model === undefined) {
      throw new ReferenceError("Unspecified model.");
    }

    if (criteria === undefined) {
      throw new ReferenceError("Unspecified criteria.");
    }

    return this.opMethods.executeQuery.call(this, model, criteria, false);
  }
  ,
  /**
   * Tests if any of the models from an array match against a criteria object
   *
   * @method some
   * @param modelArray
   * @param criteria
   * @return {Promise} who's result is set to a {Boolean}.
   *
   * @example
   *
   * ```javascript
   *  // the data to criteria
   *  var model = [{
   *    name: 'cyborg',
   *    age: 123
   *  }, {
   *    name: 'borg',
   *    age: 223
   *  }];
   *
   *  // the criteria
   *  var criteria = {
   *    age: {
   *      $equals: 223
   *    }
   *  };
   *
   *  // run the criteria
   *  var result = queryify.some(model, criteria)
   *  // the result, true or false
   * ```
   */
  some: function(modelArray, criteria) {
    var context = this,
      trace = context.trace.on;

    if (trace === true) {
      // setup the criteria logging stack
      this.trace.reset();
    }

    if (modelArray === undefined) {
      throw new ReferenceError("Unspecified model array.");
    }

    if (criteria === undefined) {
      throw new ReferenceError("Unspecified criteria.");
    }

    var result,
      traceIndex = 0,
      nextIter = modelArray[Symbol.iterator](),
      iterResult;

    // iterates the promises until true or error
    for (; !(iterResult = nextIter.next()).done;) {
      if (trace === true) {
        // log current item operation
        this.trace.stackLevel = 0;
        this.trace.write(["$$some", traceIndex++]);
      }

      result = this.opMethods.executeQuery.call(this, iterResult.value, criteria, false);
      if (result === true)
        break;
    }

    return result;
  }
};
