/*
 QueryifyJs is a No-SQL object query library made for javascript
 Copyright (C) 2014 Peter Flannery

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @module QueryifyJs
 */
"use strict";
/**
 * Default compare methods.
 *
 * @class QueryComparisons
 */
function QueryComparisons() {
}

QueryComparisons.prototype = {
  /**
   * Default compare method that is used when no compare method is specified in the criteria.
   *
   * @method $default
   * @param modelValue
   * @param testValue Can be either a string value or a {Regexp}
   * @return {Boolean}
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    author: 'cyborg',
   *    title: 'cybertron'
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    author: 'cyborg', // equiv to author: {$default: 'cyborg'}
   *    title: /^cyber/   // equiv to title: {$default: /^cyber/}
   *  };
   * ```
   */
  $default: function(modelValue, testValue) {
    if (testValue instanceof RegExp)
      return this.compareMethods.$regexp(modelValue, testValue);
    else
      return this.compareMethods.$equals(modelValue, testValue);
  },
  /**
   * Equality and type check. Equivalent to using '==='
   * @method $equals
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    name: {
   *      $equals: 'cyborg'
   *    }
   *  };
   * ```
   */
  $equals: function(modelValue, testValue) {
    return modelValue === testValue;
  },
  /**
   * Match against the length of the model value being evaluated
   * @method $length
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    name: {
   *      $length: 6
   *    }
   *  }
   * ```
   */
  $length: function(modelValue, testValue) {
    return modelValue.length == testValue;
  },
  /**
   * Regular expression
   * @method $regexp
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    age: {
   *      $regexp: /\d/ig
   *    }
   *  };
   * ```
   */
  $regexp: function(modelValue, testValue) {
    return testValue.test(modelValue)
  },
  /**
   * Match a {String} from the left
   * @method $left
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    name: {
   *      $left: 'cyb'
   *    }
   *  };
   * ```
   */
  $left: function(modelValue, testValue) {
    return modelValue.substring(0, testValue.length) === testValue;
  },
  /**
   * Matches a {String} from the right
   * @method $right
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    name: {
   *      $right: 'org'
   *    }
   *  };
   * ```
   */
  $right: function(modelValue, testValue) {
    var offset = modelValue.length - testValue.length;
    return modelValue.substring(offset) === testValue;
  },
  /**
   * Typeof
   * @method $typeof
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };

   *  // the criteria
   *  var criteria = {
   *    age: {
   *      $typeof: 'number'
   *    }
   *  };
   * ```
   */
  $typeof: function(modelValue, testValue) {
    return typeof modelValue === testValue;
  },
  /**
   * instanceof
   * @method $instanceof
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    age: {
   *      $instanceof: Object
   *    }
   *  };
   * ```
   */
  $instanceof: function(modelValue, testValue) {
    return modelValue instanceof testValue;
  },
  /**
   * Converts the modelValue to lower case before making the comparison
   * @method $toLowerCase
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'CYBORG',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    name: {
   *      $toLowerCase: 'cyborg'
   *    }
   *  };
   * ```
   */
  $toLowerCase: function(modelValue, testValue) {
    return modelValue.toLowerCase() === testValue;
  },
  /**
   * Converts the modelValue to upper case before making the comparison
   * @method $toUpperCase
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    name: {
   *      $toUpperCase: 'CYBORG'
   *    }
   *  };
   * ```
   */
  $toUpperCase: function(modelValue, testValue) {
    return modelValue.toUpperCase() === testValue;
  },
  /**
   * Matches the modelValue against an array of values
   * @method $inArray
   * @param modelValue
   * @param testArray
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *      name: 'cyborg',
   *      age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *      name: {
   *          $inArray: ['cyborg', 'or-me']
   *      }
   *  };
   * ```
   */
  $inArray: function(modelValue, testArray) {
    return testArray.indexOf(modelValue) !== -1;
  },
  /**
   * Matches the model property names against a single or an array of value(s)
   * @method $has
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *      name: 'cyborg',
   *      age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *      $has: 'name'
   *  };
   * ```
   */
  $has: function(modelValue, testValue) {
    if (testValue instanceof Array) {
      return testValue.every(function(testItemValue) {
        return (testItemValue in modelValue);
      });
    } else {
      return testValue in modelValue;
    }
  },
  /**
   * Greater than test. Equivalent to using '>'
   * @method $g
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *      name: 'cyborg',
   *      age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *      age: {
   *          $g: 100
   *      }
   *  };
   * ```
   */
  $g: function(modelValue, testValue) {
    return modelValue > testValue;
  },
  /**
   * Greater than equals test. Equivalent to using '>='
   * @method $gte
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    age: {
   *      $gte: 123
   *    }
   *  };
   * ```
   */
  $gte: function(modelValue, testValue) {
    return modelValue >= testValue;
  },
  /**
   * Less than equals test. Equivalent to using '<='
   * @method $lte
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    age: {
   *      $lte: 123
   *    }
   *  };
   * ```
   */
  $lte: function(modelValue, testValue) {
    return modelValue <= testValue;
  },
  /**
   * Less than test. Equivalent to using '<'
   * @method $l
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    age: {
   *      $l: 200
   *    }
   *  };
   * ```
   */
  $l: function(modelValue, testValue) {
    return modelValue < testValue;
  },
  /**
   * Equals test. Equivalent to using '=='
   * @method $e
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    age: {
   *      $e: 123
   *    }
   *  };
   * ```
   */
  $e: function(modelValue, testValue) {
    return modelValue == testValue;
  },
  /**
   * Not Equals test. Equivalent to using '!='
   * @method $ne
   * @param modelValue
   * @param testValue
   * @return {Boolean}
   *
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    name: 'cyborg',
   *    age: 123
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    age: {
   *      $ne: 2
   *    }
   *  };
   * ```
   */
  $ne: function(modelValue, testValue) {
    return modelValue != testValue;
  },
  /**
   * Matches a criteria against the model children properties. i.e. return queryify.one(model, criteria)
   * @method $queryOne
   * @param modelValue
   * @param criteria
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    book: {
   *      title: 'test',
   *      author: 'fred'
   *    }
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    book: {
   *      $queryOne: {
   *        title: {
   *          $equals: 'test'
   *        }
   *      }
   *    }
   *  };
   * ```
   */
  $queryOne: function(modelValue, criteria) {
    return this.opMethods.executeQuery.call(this, modelValue, criteria, false);
  },
  /**
   * Matches a criteria against an array of models. i.e. return queryify.every(modelArray, criteria)
   * @method $queryEvery
   * @param modelArray
   * @param criteria
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // the data to query
   *  var model = {
   *    books: [{
   *      title: 'test',
   *      author: 'fred'
   *    }, {
   *      title: 'another book',
   *      author: 'fred'
   *    }]
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    books: {
   *      $queryEvery: {
   *        author: {
   *          $equals: 'fred'
   *        }
   *      }
   *    }
   *  };
   * ```
   */
  $queryEvery: function(modelArray, criteria) {
    var scope = this;
    // iterates the until false or error
    return modelArray.every(function(model) {
      return scope.opMethods.executeQuery.call(scope, model, criteria, false);
    });
  },
  /**
   * Matches a criteria against an array of models. i.e. return queryify.some(modelArray, criteria)
   * @method $querySome
   * @param modelArray
   * @param criteria
   * @return {Boolean}
   * @example
   *
   * ```javascript
   *  // same as $queryEvery except iterates until finds first match, no matches or errors
   *
   *  // the data to query
   *  var model = {
   *    books: [{
   *      title: 'test',
   *      author: 'fred'
   *    }, {
   *      title: 'another book',
   *      author: 'fred'
   *    }]
   *  };
   *
   *  // the criteria
   *  var criteria = {
   *    books: {
   *      $querySome: {
   *        author: {
   *          $equals: 'fred'
   *        }
   *      }
   *  };
   * ```
   */
  $querySome: function(modelArray, criteria) {
    var scope = this;
    // iterates the until finds first match, no matches or errors
    return modelArray.some(function(model) {
      return scope.opMethods.executeQuery.call(scope, model, criteria, false);
    });
  }
};
