/*
 QueryifyJs is a No-SQL object query library made for javascript
 Copyright (C) 2014 Peter Flannery

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @module QueryifyJs
 */
/**
 * @class QueryOperations
 */
function QueryOperations() {
}

function throwError(error) {
  throw error;
}

QueryOperations.prototype = {
  /**
   * The logical AND operator (&&)
   *
   * @method $and
   * @example
   *
   * ```javascript
   *   var model = {
   *     name: 'cyborg',
   *     age: 123
   *   };
   *
   *   var criteria = {
   *     name: {
   *       $right: 'g',
   *       $and: {
   *         $left: 'cy'
   *       }
   *     },
   *     $and: {
   *       age: {
   *         $typeof: 'number',
   *         $and: {
   *           $equals: 123
   *         }
   *       }
   *     }
   *   }
   * ```
   */
  $and: function(model, operations, negateCompare, modelValue) {
    if (this.trace.on === true) {
      // log current operation
      this.trace.write('$and');
    }

    // execute the $and operation
    return this.opMethods.executeQuery.call(this, model, operations, negateCompare, modelValue);
  },
  /**
   * The logical OR operator (||)
   *
   * @method $or
   * @example
   *
   * ```javascript
   *   var model = {
   *     name: 'cyborg',
   *     age: 123
   *   };
   *
   *   var criteria = {
   *     name: {
   *       $right: 'abc',
   *       $or: {
   *         $left: 'cy'
   *       }
   *     }
   *   };
   * ```
   */
  $or: function(model, operations, negateCompare, modelValue) {
    if (this.trace.on === true) {
      // log current operation
      this.trace.write('$or');
    }
    // execute the $and operation
    return this.opMethods.executeQuery.call(this, model, operations, negateCompare, modelValue);
  },
  /**
   * The logical negation operator (!) reverses the result of all operations inside it's scope
   *
   * @method $not
   * @example
   *
   * ```javsacript
   *  var model = {
   *    name: 'cyborg',
   *      age: 123
   *  };
   *  var criteria = {
   *    $not: {
   *      name: {
   *        $typeof: 'number'
   *      },
   *      age: {
   *        $typeof: 'string'
   *      }
   *    }
   *  };
   * ```
   */
  $not: function(model, operations, modelValue) {
    var stackIndex;
    if (this.trace.on === true) {
      // store the $not stack index so we can set it's result after
      stackIndex = this.trace.stackLevel;
      // log current operation
      this.trace.write('$not');
    }

    // execute the $not operation
    var result = this.opMethods.executeQuery.call(this, model, operations, true, modelValue);
    if (this.trace.on === true) {
      // save the result to the trace stack
      this.trace.stack[stackIndex - 1].result = result;
    }

    // return the result
    return result;
  },
  executeCompare: function(operationkey, operations, negateCompare, modelValue, reject) {
    if (this.trace.on === true) {
      // log current operation
      this.trace.write(operationkey);
    }

    // check if we have a compare method
    var compareMethod = this.compareMethods[operationkey];
    if (compareMethod === undefined) {
      reject(new ReferenceError("No compare method was found for: " + operationkey));
      return;
    }

    // check if we have anything to process
    var compareValue = operations[operationkey];

    // execute the compareMethod
    var result = compareMethod.call(this, modelValue, compareValue);

    // negate the result if need be
    if (negateCompare) {
      result = !result;
    }

    if (this.trace.on === true) {
      // set the result for the current operation to the traceStack
      var traceStack = this.trace.stack;
      traceStack[traceStack.length - 1].result = result;
    }

    // return the result
    return result;
  },
  executeValueOperations: function(model, operationkey, operations, negateCompare, reject) {
    if (this.trace.on === true) {
      // log current operation
      this.trace.write(operationkey);
    }

    var descendantOps = operations[operationkey];
    if (descendantOps === undefined) {
      reject(new ReferenceError("No operation property was found for: " + operationkey));
      return;
    }

    var modelValue = model[operationkey];

    // execute the modelValue operations
    return this.opMethods.executeQuery.call(this, model, descendantOps, negateCompare, modelValue);
  },
  executeQuery: function(model, operations, negateCompare, modelValue) {
    var context = this;

    if (context.trace.on === true) {
      // increment the query level
      context.trace.stackLevel++;
    }

    // use default compare if operations is a RegExp or not an object
    if (operations instanceof RegExp || typeof operations !== 'object') {
      operations = {
        '$default': operations
      };
    }

    // loop the operation keys
    var result = false;
    for (var operationKey in operations) {
      if (operations.hasOwnProperty(operationKey)) {
        if (operationKey[0] === '$') {

          if (modelValue === undefined) {
            modelValue = model;
          }

          if (operationKey === '$not') {
            result = context.opMethods.$not.call(context, model, operations.$not, modelValue);
          } else if (operationKey === '$and') {
            result = context.opMethods.$and.call(context, model, operations.$and, negateCompare, modelValue);
          } else if (operationKey !== '$or') {
            result = context.opMethods.executeCompare.call(context,
              operationKey,
              operations,
              negateCompare,
              modelValue,
              throwError
            );
          }

        } else {
          result = context.opMethods.executeValueOperations.call(context,
            model,
            operationKey,
            operations,
            negateCompare,
            throwError
          );
        }

        if (result === false)
          break;

      }
    }

    // run the $or if present and when the result is false
    if (operations.$or && result === false) {
      result = context.opMethods.$or.call(context, model, operations.$or, negateCompare, modelValue);
    }

    if (context.trace.on === true) {
      // decrement the stack level
      context.trace.stackLevel--;
    }

    return result;
  }

};
