/*
 QueryifyJs is a No-SQL object query library made for javascript
 Copyright (C) 2014 Peter Flannery

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @module QueryifyJs
 */

/**
 * Class for creating Promise based Query instances.
 *
 * @class QueryPromise
 *
 * @constructor
 * @param {QueryOperations} queryOperations
 * @param {QueryComparisons} queryComparisons
 * @param {QueryStackTrace} queryStackTrace
 * @param {Object} sharedContext
 *      passed to the custom compare methods
 */
function QueryPromise(queryOperations, queryComparisons, queryStackTrace, sharedContext) {
  this.opMethods = queryOperations;
  this.compareMethods = queryComparisons;
  this.trace = queryStackTrace;

  // fallback to an empty object if not supplied
  this.sharedContext = sharedContext || {};
}

QueryPromise.prototype = {
  /**
   * Extracts items from an object dictionary using a criteria object
   *
   * @method extract
   * @async
   *
   * @param {Object} modelDictionary An object dictionary of model objects
   * @param {Object} criteria The criteria object
   *
   * @return an {Object} dictionary of model objects.
   *
   * @example
   *
   * ```javascript
   *   // the data to criteria
   *   var modelDictionary = {
   *     "item1": {
   *       name: 'cyborg',
   *       age: 123
   *     },
   *     "item2": {
   *       name: 'borg',
   *       age: 223
   *     }
   *   };
   *
   *   // the criteria
   *   var criteria = {
   *     name: {
   *       $equals: 'cyborg'
   *     }
   *   };
   *
   *   // run the query
   *   queryify.extract(modelDictionary, criteria)
   *     .then(function (results) {
   *       // the results, an extracted model dictionary
   *     }).catch (function (error) {
   *       // catch any errors
   *     })
   *     .then(function () {
   *       // finally
   *     });
   * ```
   */
  extract: function(modelDictionary, criteria) {
    var context = this,
      trace = context.trace.on;

    if (trace === true) {
      // setup the query logging stack
      context.trace.reset();
    }

    if (modelDictionary === undefined) {
      return Promise.reject(new ReferenceError("Unspecified model dictionary."));
    }

    if (criteria === undefined) {
      return Promise.reject(new ReferenceError("Unspecified criteria."));
    }

    // setup a chain of thenable methods
    var chainedThen = Promise.resolve({});

    var keys = Object.keys(modelDictionary);
    keys.forEach(function(modelKey, modelIndex) {

      chainedThen = chainedThen.then(function(filteredDictionary) {

        if (trace === true) {
          // log current item operation
          context.trace.stackLevel = 0;
          context.trace.write(["$$extract", modelIndex]);
        }

        var model = modelDictionary[modelKey];
        return context.opMethods.executeQuery.call(context, model, criteria, false)
          .then(function(result) {
            // determine if passed
            if (result === true) {
              filteredDictionary[modelKey] = model;
            }
            // pass on the results
            return filteredDictionary;
          });
      });

    });

    return chainedThen
  },
  /**
   * Filters an array of model objects against a criteria object
   *
   * @method filter
   * @async
   *
   * @param {Array} modelArray An array of data objects
   * @param {Object} criteria The criteria object
   *
   * @return {Promise} who's result is set to an {Array} of filtered objects.
   *
   * @example
   *
   * ```javascript
   *   // the data to query
   *   var model = [{
   *     name: 'cyborg',
   *     age: 123
   *   }, {
   *     name: 'borg',
   *     age: 223
   *   }];
   *
   *   // the criteria
   *   var criteria = {
   *     name: {
   *       $equals: 'cyborg'
   *     }
   *   };
   *
   *   // run the query
   *   queryify.filter(model, criteria)
   *     .then(function (results) {
   *       // the results, a filtered array
   *     }).catch (function (error) {
   *       // catch any errors
   *     })
   *     .then(function () {
   *       // finally
   *     });
   * ```
   */
  filter: function(modelArray, criteria) {
    var context = this,
      trace = context.trace.on;

    if (trace === true) {
      // setup the query logging stack
      context.trace.reset();
    }

    if (modelArray === undefined) {
      return Promise.reject(new ReferenceError("Unspecified model array."));
    }

    if (criteria === undefined) {
      return Promise.reject(new ReferenceError("Unspecified criteria."));
    }

    // create a chain of thenables
    var chainedThen = Promise.resolve([]);

    modelArray.forEach(function(model, modelIndex) {

      chainedThen = chainedThen.then(function(filteredList) {
        if (trace === true) {
          // log current operation
          context.trace.stackLevel = 0;
          context.trace.write(["$$filter", modelIndex]);
        }

        return context.opMethods.executeQuery.call(context, model, criteria, false)
          .then(function(result) {
            // determine if passed
            if (result === true) {
              filteredList.push(model);
            }
            // pass on the results
            return filteredList;
          });

      });

    });

    return chainedThen;
  },
  /**
   * Tests if an array of model objects match against a criteria object
   *
   * @method every
   * @async
   *
   * @param {Array} modelArray An array of data objects
   * @param {Object} criteria The criteria object
   *
   * @return {Promise} who's result is set to a {Boolean}.
   *
   * @example
   *
   * ```javascript
   *   // the data to query
   *   var model = [{
   *     name: 'cyborg',
   *     age: 123
   *   }, {
   *     name: 'borg',
   *     age: 223
   *   }];
   *
   *   // the criteria
   *   var criteria = {
   *     name: {
   *       $right: 'g'
   *     },
   *     age: {
   *       $typeof: 'number'
   *     }
   *   };
   *
   *   // run the criteria
   *   queryify.every(model, criteria)
   *     .then(function (result) {
   *       // the result, true or false
   *     })
   *     .catch (function (error) {
   *       // catch any errors
   *     })
   *     .then(function () {
   *       // finally
   *     });
   * ```
   */
  every: function(modelArray, criteria) {
    var context = this,
      trace = context.trace.on;

    if (trace === true) {
      // setup the query logging stack
      context.trace.reset();
    }

    if (modelArray === undefined) {
      return Promise.reject(new ReferenceError("Unspecified model array."));
    }

    if (criteria === undefined) {
      return Promise.reject(new ReferenceError("Unspecified criteria."));
    }

    // iterates the promises until false or error
    return promiseExtras.every(modelArray, function(model, modelIndex, fulfil) {
      if (trace === true) {
        // log current item operation
        context.trace.stackLevel = 0;
        context.trace.write(["$$every", modelIndex]);
      }

      fulfil(context.opMethods.executeQuery.call(context, model, criteria, false));
    });
  },
  /**
   * Tests one model object against a criteria object
   *
   * @method one
   * @async
   *
   * @param {Object} model The data object
   * @param {Object} criteria The criteria object
   *
   * @return {Promise} who's result is set to a {Boolean}.
   *
   * @example
   *
   * ```javascript
   *   // the data to query
   *   var model = {
   *     name: 'cyborg',
   *     age: 123
   *   };
   *
   *   // the criteria
   *   var criteria = {
   *     name: {
   *       $right: 'g',
   *       $and: {
   *         $left: 'cy'
   *       }
   *     },
   *     $and: {
   *       age: {
   *         $typeof: 'number',
   *           $and: {
   *             $equals: 123
   *           }
   *       }
   *     }
   *   };
   *
   *   // run the criteria
   *   queryify.one(model, criteria)
   *     .then(function (result) {
   *       // the result
   *     }).catch (function (error) {
   *       // catch any errors
   *     })
   *     .then(function () {
   *       // finally
   *     });
   * ```
   */
  one: function(model, criteria) {
    if (this.trace.on === true) {
      // setup the query logging stack
      this.trace.reset();
    }

    if (model === undefined) {
      return Promise.reject(new ReferenceError("Unspecified model."));
    }

    if (criteria === undefined) {
      return Promise.reject(new ReferenceError("Unspecified criteria."));
    }

    return this.opMethods.executeQuery.call(this, model, criteria, false);
  },
  /**
   * Tests if any of the models from an array match against a criteria object
   *
   * @method some
   * @param modelArray
   * @param criteria
   * @return {Promise} who's result is set to a {Boolean}.
   *
   * @example
   *
   * ```javascript
   *   // the data to query
   *   var model = [{
   *     name: 'cyborg',
   *     age: 123
   *   }, {
   *     name: 'borg',
   *     age: 223
   *   }];
   *
   *   // the criteria
   *   var criteria = {
   *     age: {
   *       $equals: 223
   *     }
   *   };
   *
   *   // run the criteria
   *   queryify.some(model, criteria)
   *     .then(function (result) {
   *       // the result, true or false
   *     })
   *     .catch (function (error) {
   *       // catch any errors
   *     })
   *     .then(function () {
   *       // finally
   *     });
   * ```
   */
  some: function(modelArray, criteria) {
    var context = this,
      trace = context.trace.on;

    if (trace === true) {
      // setup the query logging stack
      context.trace.reset();
    }

    if (modelArray === undefined) {
      return Promise.reject(new ReferenceError("Unspecified model array."));
    }

    if (criteria === undefined) {
      return Promise.reject(new ReferenceError("Unspecified criteria."));
    }

    // iterates the promises until false or error
    return promiseExtras.some(modelArray, function(model, modelIndex, fulfil) {
      if (trace === true) {
        // log current item operation
        context.trace.stackLevel = 0;
        context.trace.write(["$$some", modelIndex]);
      }
      fulfil(context.opMethods.executeQuery.call(context, model, criteria, false));
    });
  }
};
