/*
 QueryifyJs is a No-SQL object query library made for javascript
 Copyright (C) 2014 Peter Flannery

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * ###A No-SQL object query library made for javascript
 *
 * ####Installing:
 *
 * ######NodeJs:
 *
 * Will automatically reference any dependencies.
 *
 *     npm install querifyjs --save
 *
 * ######Bower:
 *
 *     bower install querifyjs --save
 *
 * ####Referencing:
 *
 * ######Browser:
 *
 * Dependencies:
 *
 *   [promise-extras](http://github.com/pflannery/promise-extras) Only when you need to use promises, otherwise not required
 *
 *   [dist/queryify.all.min.js](https://github.com/pflannery/querifyjs/) includes this dependency
 *
 *     <script type="application/javascript" src="path/to/lib/queryify.all.min.js"></script>
 *
 * otherwise manually place before queryify library i.e.
 *
 *     <script type="application/javascript" src="path/to/lib/promise-extras.min.js"></script>
 *     <script type="application/javascript" src="path/to/lib/queryify.min.js"></script>
 *
 *  ```javascript
 *  // global Query instance
 *  var queryify = window.queryify.sync;
 *
 *  // global QueryPromise instance (lazy loads deps on first access to this property)
 *  var queryify = window.queryify.promise;
 *  ```
 *
 * ######NodeJs:
 *
 *  ```javascript
 *  // global Query instance (for synchronous operations)
 *  var queryify = require('querifyjs').sync;
 *
 *  // global QueryPromise instance
 *  var queryify = require('querifyjs').promise;
 *  ```
 *
 * ######Overview Examples:
 *
 *  ```javascript
 *  var models = [
 *    {name: 'apple'},
 *    {name: 'banana'},
 *    {name: 'pineapple'},
 *    {name: 'strawberry'},
 *    {name: 'orange'},
 *    {name: 'grapefruit'}
 *  ];
 *
 *  var criteria = {
 *    name: /apple/,      // regexp to match any names with 'apple' in them
 *    $or: {              // or
 *      name: 'banana'  // match name against 'banana'
 *    }
 *  };
 *
 *  var results = queryify.sync.filter(models, criteria);
 *  // produces [{"name":"apple"},{"name":"banana"},{"name":"pineapple"}]
 *
 *  var criteria = {
 *    name: {
 *      $right: 'e'     // match the last char in name that contains 'e'
 *    }
 *  };
 *
 *  var results = queryify.sync.filter(models, criteria);
 *  // produces [{"name":"apple"},{"name":"pineapple"},{"name":"orange"}]
 *
 *  // direct value queries
 *
 *  var result = queryify.sync.one("myvalue", {$equals: 'myvalue'});
 *  // promise based example
 *
 *  var criteria = {
 *    name: /an/          // regexp to match any names with 'an'
 *  };
 *
 *  queryify.promise.filter(models, criteria)
 *    .then(function(results) {
 *      // produces [{"name":"banana"},{"name":"orange"}]
 *    );
 *  ```
 *
 * @module QueryifyJs
 * @main QueryifyJs
 */
var isBrowser = typeof window !== typeof undefined,
  $Promise,
  $promiseExtras;

function resolveBrowserPromiseLib() {
  if (window.Promise) {
    return window.Promise;
  } else {
    throw new ReferenceError("Promise library could not be detected for this browser.");
  }
}

function resolveBrowserPromiseExtrasLib() {
  if (window.promiseExtras) {
    return window.promiseExtras;
  } else {
    throw new ReferenceError("Promise-extras library could not be detected for this browser.");
  }
}

function lazyLoadPromiseDeps() {
  // attempt to get the Promise ref
  if (isBrowser === true) {
    $Promise = resolveBrowserPromiseLib();
  } else {
    $Promise = require("promise");
  }

  // attempt to get the Promise ref
  if (isBrowser === true) {
    $promiseExtras = resolveBrowserPromiseExtrasLib();
  } else {
    $promiseExtras = require("promise-extras");
  }
}

/**
 * global object
 *
 * @class Global
 *
 */
var defaultPromiseInstance;
var queryifyLibrary = {
  /**
   * A reference to the global QueryPromise instance.
   *
   * @property {QueryPromise} promise
   * @example
   *
   *  ```javascript
   *  // referencing the Promise instance
   *  var queryify = window.queryify.promise;
   *
   *  // accessing the interface
   *  queryify.one(model, criteria)
   *   .then(function(result) {
   *
   *   });
   *  ```
   */
  get promise() {
    if (defaultPromiseInstance === undefined) {
      lazyLoadPromiseDeps();
      defaultPromiseInstance = new QueryPromise(
        new QueryPromiseOperations(),
        new QueryPromiseComparisons(),
        new QueryStackTrace()
      );
    }
    return defaultPromiseInstance;
  },
  /**
   * A reference to the synchronous Query global instance.
   *
   * @property {Query} sync
   * @example
   *
   *  ```javascript
   *  // referencing the synchronous instance
   *  var queryify = window.queryify.sync;
   *
   *  // accessing the interface
   *  queryify.one(model, criteria);
   *  ```
   */
  sync: new Query(new QueryOperations(), new QueryComparisons(), new QueryStackTrace()),

  /**
   * A reference to the Query class type for custom initiation.
   * @property {Query} Query
   */
  Query: Query,
  /**
   * A reference to the QueryOperations class type for custom initiation.
   * @property {QueryOperations} QueryOperations
   */
  QueryOperations: QueryOperations,
  /**
   * A reference to the QueryComparisons class type for custom initiation.
   * @property {QueryComparisons} QueryComparisons
   */
  QueryComparisons: QueryComparisons,
  /**
   * A reference to the QueryStackTrace class type for custom initiation.
   * @property {QueryStackTrace} QueryStackTrace
   */
  QueryStackTrace: QueryStackTrace,

  // expose the promise classes (lazy loads when accessed)
  /**
   * An reference to the QueryPromise class type for custom initiation.
   * @property {QueryPromise} QueryPromise
   */
  get QueryPromise() {
    if (defaultPromiseInstance === undefined)
      lazyLoadPromiseDeps();
    return QueryPromise;
  },
  /**
   * An reference to the QueryPromiseOperations class type for custom initiation.
   * @property {QueryPromiseOperations} QueryPromiseOperations
   */
  get QueryPromiseOperations() {
    if (defaultPromiseInstance === undefined)
      lazyLoadPromiseDeps();
    return QueryPromiseOperations;
  },
  /**
   * An reference to the QueryPromiseComparisons class type for custom initiation.
   * @property {QueryPromiseComparisons} QueryPromiseComparisons
   */
  get QueryPromiseComparisons() {
    if (defaultPromiseInstance === undefined)
      lazyLoadPromiseDeps();
    return QueryPromiseComparisons;
  },

  get polyfills() {
    return {
      ArrayValuesIterator: ArrayValuesIterator,
      ArrayEntriesIterator: ArrayEntriesIterator
    };
  }

};

/* export the library */
if (typeof module !== typeof undefined && typeof module.exports !== typeof undefined) {
  /* node */
  module.exports = queryifyLibrary;
} else {
  var indirectEval = (0, eval),
    globalThis = indirectEval('this');

  globalThis.queryify = queryifyLibrary;
}
